@@ -29,6 +29,12 @@ This plugin checks a message against Sender Policy Framework (SPF)
 records published by the domain owners in DNS to fight email address
 forgery and make it easier to identify spams.
 
+It's recommended to use MTA filter (pypolicyd-spf / spf-engine etc), so this
+plugin can reuse the Received-SPF and/or Authentication-Results header results as is.
+Otherwise throughput could suffer, DNS lookups done by this plugin are not
+asynchronous.
+Those headers will also help when SpamAssassin is not able to correctly detect EnvelopeFrom.
+
 =cut
 
 package Mail::SpamAssassin::Plugin::SPF;
@@ -53,22 +59,25 @@ sub new {
   my $self = $class->SUPER::new($mailsaobject);
   bless ($self, $class);
 
-  $self->register_eval_rule ("check_for_spf_pass");
-  $self->register_eval_rule ("check_for_spf_neutral");
-  $self->register_eval_rule ("check_for_spf_none");
-  $self->register_eval_rule ("check_for_spf_fail");
-  $self->register_eval_rule ("check_for_spf_softfail");
-  $self->register_eval_rule ("check_for_spf_permerror");
-  $self->register_eval_rule ("check_for_spf_temperror");
-  $self->register_eval_rule ("check_for_spf_helo_pass");
-  $self->register_eval_rule ("check_for_spf_helo_neutral");
-  $self->register_eval_rule ("check_for_spf_helo_none");
-  $self->register_eval_rule ("check_for_spf_helo_fail");
-  $self->register_eval_rule ("check_for_spf_helo_softfail");
-  $self->register_eval_rule ("check_for_spf_helo_permerror");
-  $self->register_eval_rule ("check_for_spf_helo_temperror");
-  $self->register_eval_rule ("check_for_spf_whitelist_from");
-  $self->register_eval_rule ("check_for_def_spf_whitelist_from");
+  $self->register_eval_rule ("check_for_spf_pass", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_neutral", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_none", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_fail", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_softfail", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_permerror", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_temperror", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_helo_pass", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_helo_neutral", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_helo_none", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_helo_fail", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_helo_softfail", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_helo_permerror", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_helo_temperror", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_welcomelist_from", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_spf_whitelist_from", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS); # removed in 4.1
+  $self->register_eval_rule ("check_for_def_spf_welcomelist_from", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+  $self->register_eval_rule ("check_for_def_spf_whitelist_from", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS); # removed in 4.1
+  $self->register_eval_rule ("check_spf_skipped_noenvfrom", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
 
   $self->set_config($mailsaobject->{conf});
 
@@ -85,37 +94,43 @@ sub set_config {
 
 =over 4
 
-=item whitelist_from_spf user@example.com
+=item welcomelist_from_spf user@example.com
 
-Works similarly to whitelist_from, except that in addition to matching
-a sender address, a check against the domain's SPF record must pass.
-The first parameter is an address to whitelist, and the second is a string
-to match the relay's rDNS.
+Previously whitelist_from_spf which will work interchangeably until 4.1.
 
-Just like whitelist_from, multiple addresses per line, separated by spaces,
-are OK. Multiple C<whitelist_from_spf> lines are also OK.
+Works similarly to welcomelist_from, except that in addition to matching a
+sender address, a check against the domain's SPF record must pass.  The
+first parameter is an address to welcomelist, and the second is a string to
+match the relay's rDNS.
 
-The headers checked for whitelist_from_spf addresses are the same headers
+Just like welcomelist_from, multiple addresses per line, separated by
+spaces, are OK.  Multiple C<welcomelist_from_spf> lines are also OK.
+
+The headers checked for welcomelist_from_spf addresses are the same headers
 used for SPF checks (Envelope-From, Return-Path, X-Envelope-From, etc).
 
-Since this whitelist requires an SPF check to be made, network tests must be
+Since this welcomelist requires an SPF check to be made, network tests must be
 enabled. It is also required that your trust path be correctly configured.
 See the section on C<trusted_networks> for more info on trust paths.
 
 e.g.
 
-  whitelist_from_spf joe@example.com fred@example.com
-  whitelist_from_spf *@example.com
+  welcomelist_from_spf joe@example.com fred@example.com
+  welcomelist_from_spf *@example.com
 
-=item def_whitelist_from_spf user@example.com
+=item def_welcomelist_from_spf user@example.com
 
-Same as C<whitelist_from_spf>, but used for the default whitelist entries
-in the SpamAssassin distribution.  The whitelist score is lower, because
+Previously def_whitelist_from_spf which will work interchangeably until 4.1.
+
+Same as C<welcomelist_from_spf>, but used for the default welcomelist entries
+in the SpamAssassin distribution.  The welcomelist score is lower, because
 these are often targets for spammer spoofing.
 
-=item unwhitelist_from_spf user@example.com
+=item unwelcomelist_from_spf user@example.com
 
-Used to remove a C<whitelist_from_spf> or C<def_whitelist_from_spf> entry. 
+Previously unwhitelist_from_spf which will work interchangeably until 4.1.
+
+Used to remove a C<welcomelist_from_spf> or C<def_welcomelist_from_spf> entry. 
 The specified email address has to match exactly the address previously used.
 
 Useful for removing undesired default entries from a distributed configuration
@@ -124,17 +139,20 @@ by a local or site-specific configuration or by C<user_prefs>.
 =cut
 
   push (@cmds, {
-    setting => 'whitelist_from_spf',
+    setting => 'welcomelist_from_spf',
+    aliases => ['whitelist_from_spf'], # removed in 4.1
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_ADDRLIST
   });
 
   push (@cmds, {
-    setting => 'def_whitelist_from_spf',
+    setting => 'def_welcomelist_from_spf',
+    aliases => ['def_whitelist_from_spf'], # removed in 4.1
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_ADDRLIST
   });
 
   push (@cmds, {
-    setting => 'unwhitelist_from_spf',
+    setting => 'unwelcomelist_from_spf',
+    aliases => ['unwhitelist_from_spf'], # removed in 4.1
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_ADDRLIST,
     code => sub {
       my ($self, $key, $value, $line) = @_;
@@ -144,9 +162,9 @@ by a local or site-specific configuration or by C<user_prefs>.
       unless ($value =~ /^(?:\S+(?:\s+\S+)*)$/) {
         return $Mail::SpamAssassin::Conf::INVALID_VALUE;
       }
-      $self->{parser}->remove_from_addrlist('whitelist_from_spf',
+      $self->{parser}->remove_from_addrlist('welcomelist_from_spf',
                                         split (/\s+/, $value));
-      $self->{parser}->remove_from_addrlist('def_whitelist_from_spf',
+      $self->{parser}->remove_from_addrlist('def_welcomelist_from_spf',
                                         split (/\s+/, $value));
     }
   });
@@ -173,38 +191,6 @@ days, weeks).
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_DURATION
   });
 
-=item do_not_use_mail_spf (0|1)		(default: 0)
-
-By default the plugin will try to use the Mail::SPF module for SPF checks if
-it can be loaded.  If Mail::SPF cannot be used the plugin will fall back to
-using the legacy Mail::SPF::Query module if it can be loaded.
-
-Use this option to stop the plugin from using Mail::SPF and cause it to try to
-use Mail::SPF::Query instead.
-
-=cut
-
-  push(@cmds, {
-    setting => 'do_not_use_mail_spf',
-    is_admin => 1,
-    default => 0,
-    type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
-  });
-
-=item do_not_use_mail_spf_query (0|1)	(default: 0)
-
-As above, but instead stop the plugin from trying to use Mail::SPF::Query and
-cause it to only try to use Mail::SPF.
-
-=cut
-
-  push(@cmds, {
-    setting => 'do_not_use_mail_spf_query',
-    is_admin => 1,
-    default => 0,
-    type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
-  });
-
 =item ignore_received_spf_header (0|1)	(default: 0)
 
 By default, to avoid unnecessary DNS lookups, the plugin will try to use the
@@ -250,35 +236,64 @@ working downwards until results are successfully parsed.
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
   });
 
+  # Deprecated since 4.0.0, leave for backwards compatibility
+  push(@cmds, {
+    setting => 'do_not_use_mail_spf',
+    is_admin => 1,
+    default => 0,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
+  });
+  push(@cmds, {
+    setting => 'do_not_use_mail_spf_query',
+    is_admin => 1,
+    default => 1,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
+  });
+
   $conf->{parser}->register_commands(\@cmds);
 }
 
-
 =item has_check_for_spf_errors
 
 Adds capability check for "if can()" for check_for_spf_permerror, check_for_spf_temperror, check_for_spf_helo_permerror and check_for_spf_helo_permerror
 
-=cut 
+=cut
 
 sub has_check_for_spf_errors { 1 }
 
+=item has_check_spf_skipped_noenvfrom
+
+Adds capability check for "if can()" for check_spf_skipped_noenvfrom
+
+=cut
+
+sub has_check_spf_skipped_noenvfrom { 1 }
+
+sub parsed_metadata {
+  my ($self, $opts) = @_;
+
+  $self->_get_sender($opts->{permsgstatus});
+
+  return 1;
+}
+
 # SPF support
 sub check_for_spf_pass {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 0) unless $scanner->{spf_checked};
-  $scanner->{spf_pass};
+  return $scanner->{spf_pass} ? 1 : 0;
 }
 
 sub check_for_spf_neutral {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 0) unless $scanner->{spf_checked};
-  $scanner->{spf_neutral};
+  return $scanner->{spf_neutral} ? 1 : 0;
 }
 
 sub check_for_spf_none {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 0) unless $scanner->{spf_checked};
-  $scanner->{spf_none};
+  return $scanner->{spf_none} ? 1 : 0;
 }
 
 sub check_for_spf_fail {
@@ -287,43 +302,43 @@ sub check_for_spf_fail {
   if ($scanner->{spf_failure_comment}) {
     $scanner->test_log ($scanner->{spf_failure_comment});
   }
-  $scanner->{spf_fail};
+  return $scanner->{spf_fail} ? 1 : 0;
 }
 
 sub check_for_spf_softfail {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 0) unless $scanner->{spf_checked};
-  $scanner->{spf_softfail};
+  return $scanner->{spf_softfail} ? 1 : 0;
 }
 
 sub check_for_spf_permerror {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 0) unless $scanner->{spf_checked};
-  $scanner->{spf_permerror};
+  return $scanner->{spf_permerror} ? 1 : 0;
 }
 
 sub check_for_spf_temperror {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 0) unless $scanner->{spf_checked};
-  $scanner->{spf_temperror};
+  return $scanner->{spf_temperror} ? 1 : 0;
 }
 
 sub check_for_spf_helo_pass {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 1) unless $scanner->{spf_helo_checked};
-  $scanner->{spf_helo_pass};
+  return $scanner->{spf_helo_pass} ? 1 : 0;
 }
 
 sub check_for_spf_helo_neutral {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 1) unless $scanner->{spf_helo_checked};
-  $scanner->{spf_helo_neutral};
+  return $scanner->{spf_helo_neutral} ? 1 : 0;
 }
 
 sub check_for_spf_helo_none {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 1) unless $scanner->{spf_helo_checked};
-  $scanner->{spf_helo_none};
+  return $scanner->{spf_helo_none} ? 1 : 0;
 }
 
 sub check_for_spf_helo_fail {
@@ -332,38 +347,60 @@ sub check_for_spf_helo_fail {
   if ($scanner->{spf_helo_failure_comment}) {
     $scanner->test_log ($scanner->{spf_helo_failure_comment});
   }
-  $scanner->{spf_helo_fail};
+  return $scanner->{spf_helo_fail} ? 1 : 0;
 }
 
 sub check_for_spf_helo_softfail {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 1) unless $scanner->{spf_helo_checked};
-  $scanner->{spf_helo_softfail};
+  return $scanner->{spf_helo_softfail} ? 1 : 0;
 }
 
 sub check_for_spf_helo_permerror {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 1) unless $scanner->{spf_helo_checked};
-  $scanner->{spf_helo_permerror};
+  return $scanner->{spf_helo_permerror} ? 1 : 0;
 }
 
 sub check_for_spf_helo_temperror {
   my ($self, $scanner) = @_;
   $self->_check_spf ($scanner, 1) unless $scanner->{spf_helo_checked};
-  $scanner->{spf_helo_temperror};
+  return $scanner->{spf_helo_temperror} ? 1 : 0;
 }
 
-sub check_for_spf_whitelist_from {
+=over 4
+
+=item check_spf_skipped_noenvfrom
+
+Checks if SPF checks have been skipped because EnvelopeFrom cannot be determined.
+
+=back
+
+=cut
+
+sub check_spf_skipped_noenvfrom {
   my ($self, $scanner) = @_;
-  $self->_check_spf_whitelist($scanner) unless $scanner->{spf_whitelist_from_checked};
-  $scanner->{spf_whitelist_from};
+  $self->_check_spf ($scanner, 0) unless $scanner->{spf_checked};
+  if (!exists $scanner->{spf_sender}) {
+    return 1;
+  } else {
+    return 0;
+  }
 }
 
-sub check_for_def_spf_whitelist_from {
+sub check_for_spf_welcomelist_from {
   my ($self, $scanner) = @_;
-  $self->_check_def_spf_whitelist($scanner) unless $scanner->{def_spf_whitelist_from_checked};
-  $scanner->{def_spf_whitelist_from};
+  $self->_check_spf_welcomelist($scanner) unless $scanner->{spf_welcomelist_from_checked};
+  return $scanner->{spf_welcomelist_from} ? 1 : 0;
 }
+*check_for_spf_whitelist_from = \&check_for_spf_welcomelist_from; # removed in 4.1
+
+sub check_for_def_spf_welcomelist_from {
+  my ($self, $scanner) = @_;
+  $self->_check_def_spf_welcomelist($scanner) unless $scanner->{def_spf_welcomelist_from_checked};
+  return $scanner->{def_spf_welcomelist_from} ? 1 : 0;
+}
+*check_for_def_spf_whitelist_from = \&check_for_def_spf_welcomelist_from; # removed in 4.1
 
 sub _check_spf {
   my ($self, $scanner, $ishelo) = @_;
@@ -387,7 +424,7 @@ sub _check_spf {
     $scanner->{checked_for_received_spf_header} = 1;
     dbg("spf: checking to see if the message has a Received-SPF header that we can use");
 
-    my @internal_hdrs = split("\n", $scanner->get('ALL-INTERNAL'));
+    my @internal_hdrs = $scanner->get('ALL-INTERNAL');
     unless ($scanner->{conf}->{use_newest_received_spf_header}) {
       # look for the LAST (earliest in time) header, it'll be the most accurate
       @internal_hdrs = reverse(@internal_hdrs);
@@ -459,7 +496,7 @@ sub _check_spf {
 	  dbg("spf: could not parse result from existing Received-SPF header");
 	}
 
-      } elsif ($hdr =~ /^Authentication-Results:.*;\s*SPF\s*=\s*([^;]*)/i) {
+      } elsif ($hdr =~ /^(?:Arc\-)?Authentication-Results:.*;\s*SPF\s*=\s*([^;]*)/i) {
         dbg("spf: found an Authentication-Results header added by an internal host: $hdr");
 
         # RFC 5451 header parser - added by D. Stussy 2010-09-09:
@@ -524,8 +561,6 @@ sub _check_spf {
   unless (defined $self->{has_mail_spf}) {
     my $eval_stat;
     eval {
-      die("Mail::SPF disabled by admin setting\n") if $scanner->{conf}->{do_not_use_mail_spf};
-
       require Mail::SPF;
       if (!defined $Mail::SPF::VERSION || $Mail::SPF::VERSION < 2.001) {
 	die "Mail::SPF 2.001 or later required, this is ".
@@ -547,43 +582,18 @@ sub _check_spf {
       dbg("spf: using Mail::SPF for SPF checks");
       $self->{has_mail_spf} = 1;
     } else {
-      # strip the @INC paths... users are going to see it and think there's a problem even though
-      # we're going to fall back to Mail::SPF::Query (which will display the same paths if it fails)
-      $eval_stat =~ s#^Can't locate Mail/SPFd.pm in \@INC .*#Can't locate Mail/SPFd.pm#;
-      dbg("spf: cannot load Mail::SPF module or create Mail::SPF::Server object: $eval_stat");
-      dbg("spf: attempting to use legacy Mail::SPF::Query module instead");
-
-      undef $eval_stat;
-      eval {
-	die("Mail::SPF::Query disabled by admin setting\n") if $scanner->{conf}->{do_not_use_mail_spf_query};
-
-	require Mail::SPF::Query;
-	if (!defined $Mail::SPF::Query::VERSION || $Mail::SPF::Query::VERSION < 1.996) {
-	  die "Mail::SPF::Query 1.996 or later required, this is ".
-	    (defined $Mail::SPF::Query::VERSION ? $Mail::SPF::Query::VERSION : 'unknown')."\n";
-	}
-        1;
-      } or do {
-        $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
-      };
-
-      if (!defined($eval_stat)) {
-	dbg("spf: using Mail::SPF::Query for SPF checks");
-	$self->{has_mail_spf} = 0;
-      } else {
-	dbg("spf: cannot load Mail::SPF::Query module: $eval_stat");
-	dbg("spf: one of Mail::SPF or Mail::SPF::Query is required for SPF checks, SPF checks disabled");
-	$self->{no_spf_module} = 1;
-	return;
-      }
+      dbg("spf: cannot load Mail::SPF: module: $eval_stat");
+      dbg("spf: Mail::SPF is required for SPF checks, SPF checks disabled");
+      $self->{no_spf_module} = 1;
+      return;
     }
   }
 
-
   # skip SPF checks if the A/MX records are nonexistent for the From
   # domain, anyway, to avoid crappy messages from slowing us down
   # (bug 3016)
-  return if $scanner->check_for_from_dns();
+  # TODO: this will only work if the queries are ready before SPF, so never?
+  return if $scanner->{sender_host_fail} && $scanner->{sender_host_fail} == 2;
 
   if ($ishelo) {
     # SPF HELO-checking variant
@@ -609,7 +619,7 @@ sub _check_spf {
     $scanner->{spf_failure_comment} = undef;
   }
 
-  my $lasthop = $self->_get_relay($scanner);
+  my $lasthop = $scanner->{relays_external}->[0];
   if (!defined $lasthop) {
     dbg("spf: no suitable relay for spf use found, skipping SPF%s check",
         $ishelo ? '-helo' : '');
@@ -618,7 +628,6 @@ sub _check_spf {
 
   my $ip = $lasthop->{ip};	# always present
   my $helo = $lasthop->{helo};	# could be missing
-  $scanner->{sender} = '' unless $scanner->{sender_got};
 
   if ($ishelo) {
     unless ($helo) {
@@ -627,19 +636,17 @@ sub _check_spf {
     }
     dbg("spf: checking HELO (helo=$helo, ip=$ip)");
   } else {
-    $self->_get_sender($scanner) unless $scanner->{sender_got};
-
     # TODO: we're supposed to use the helo domain as the sender identity (for
     # mfrom checks) if the sender is the null sender, however determining that
     # it's the null sender, and not just a failure to get the envelope isn't
     # exactly trivial... so for now we'll just skip the check
 
-    if (!$scanner->{sender}) {
+    if (!$scanner->{spf_sender}) {
       # we already dbg'd that we couldn't get an Envelope-From and can't do SPF
       return;
     }
     dbg("spf: checking EnvelopeFrom (helo=%s, ip=%s, envfrom=%s)",
-        ($helo ? $helo : ''), $ip, $scanner->{sender});
+        ($helo ? $helo : ''), $ip, $scanner->{spf_sender});
   }
 
   # this test could probably stand to be more strict, but try to test
@@ -657,81 +664,39 @@ sub _check_spf {
 
   my ($result, $comment, $text, $err);
 
-  # use Mail::SPF if it was available, otherwise use the legacy Mail::SPF::Query
-  if ($self->{has_mail_spf}) {
+  # TODO: currently we won't get to here for a mfrom check with a null sender
+  my $identity = $ishelo ? $helo : ($scanner->{spf_sender}); # || $helo);
 
-    # TODO: currently we won't get to here for a mfrom check with a null sender
-    my $identity = $ishelo ? $helo : ($scanner->{sender}); # || $helo);
+  unless ($identity) {
+    dbg("spf: cannot determine %s identity, skipping %s SPF check",
+        ($ishelo ? 'helo' : 'mfrom'),  ($ishelo ? 'helo' : 'mfrom') );
+    return;
+  }
+  $helo ||= 'unknown';  # only used for macro expansion in the mfrom explanation
 
-    unless ($identity) {
-      dbg("spf: cannot determine %s identity, skipping %s SPF check",
-          ($ishelo ? 'helo' : 'mfrom'),  ($ishelo ? 'helo' : 'mfrom') );
-      return;
-    }
-    $helo ||= 'unknown';  # only used for macro expansion in the mfrom explanation
+  my $request;
+  eval {
+    $request = Mail::SPF::Request->new( scope => $ishelo ? 'helo' : 'mfrom',
+			  identity      => $identity,
+			  ip_address    => $ip,
+			  helo_identity => $helo );
+    1;
+  } or do {
+    my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+    dbg("spf: cannot create Mail::SPF::Request object: $eval_stat");
+    return;
+  };
 
-    my $request;
-    eval {
-      $request = Mail::SPF::Request->new( scope         => $ishelo ? 'helo' : 'mfrom',
-					  identity      => $identity,
-					  ip_address    => $ip,
-					  helo_identity => $helo );
-      1;
-    } or do {
-      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
-      dbg("spf: cannot create Mail::SPF::Request object: $eval_stat");
-      return;
-    };
+  my $timeout = $scanner->{conf}->{spf_timeout};
 
-    my $timeout = $scanner->{conf}->{spf_timeout};
-
-    my $timer = Mail::SpamAssassin::Timeout->new(
-                { secs => $timeout, deadline => $scanner->{master_deadline} });
-    $err = $timer->run_and_catch(sub {
-
-      my $query = $self->{spf_server}->process($request);
-
-      $result = $query->code;
-      $comment = $query->authority_explanation if $query->can("authority_explanation");
-      $text = $query->text;
-
-    });
-
-
-  } else {
-
-    if (!$helo) {
-      dbg("spf: cannot get HELO, cannot use Mail::SPF::Query, consider installing Mail::SPF");
-      return;
-    }
-
-    # TODO: if we start doing checks on the null sender using the helo domain
-    # be sure to fix this so that it uses the correct sender identity
-    my $query;
-    eval {
-      $query = Mail::SPF::Query->new (ip => $ip,
-				    sender => $scanner->{sender},
-				    helo => $helo,
-				    debug => 0,
-				    trusted => 0);
-      1;
-    } or do {
-      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
-      dbg("spf: cannot create Mail::SPF::Query object: $eval_stat");
-      return;
-    };
-
-    my $timeout = $scanner->{conf}->{spf_timeout};
-
-    my $timer = Mail::SpamAssassin::Timeout->new(
-                { secs => $timeout, deadline => $scanner->{master_deadline} });
-    $err = $timer->run_and_catch(sub {
-
-      ($result, $comment) = $query->result();
-
-    });
-
-  } # end of differences between Mail::SPF and Mail::SPF::Query
+  my $timer_spf = Mail::SpamAssassin::Timeout->new(
+              { secs => $timeout, deadline => $scanner->{master_deadline} });
+  $err = $timer_spf->run_and_catch(sub {
+    my $query = $self->{spf_server}->process($request);
+    $result = $query->code;
+    $comment = $query->authority_explanation if $query->can("authority_explanation");
+    $text = $query->text;
+  });
 
   if ($err) {
     chomp $err;
@@ -739,7 +704,6 @@ sub _check_spf {
     return 0;
   }
 
-
   $result ||= 'timeout';	# bug 5077
   $comment ||= '';
   $comment =~ s/\s+/ /gs;	# no newlines please
@@ -774,138 +738,131 @@ sub _check_spf {
     }
   }
 
-  dbg("spf: query for $scanner->{sender}/$ip/$helo: result: $result, comment: $comment, text: $text");
-}
-
-sub _get_relay {
-  my ($self, $scanner) = @_;
-
-  # dos: first external relay, not first untrusted
-  return $scanner->{relays_external}->[0];
+  if ($ishelo) {
+    dbg("spf: query for $ip/$helo: result: $result, comment: $comment, text: $text");
+  } else {
+    dbg("spf: query for $scanner->{spf_sender}/$ip/$helo: result: $result, comment: $comment, text: $text");
+  }
 }
 
 sub _get_sender {
   my ($self, $scanner) = @_;
-  my $sender;
 
-  $scanner->{sender_got} = 1;
-  $scanner->{sender} = '';
-
-  my $relay = $self->_get_relay($scanner);
+  my $relay = $scanner->{relays_external}->[0];
   if (defined $relay) {
-    $sender = $relay->{envfrom};
+    my $sender = $relay->{envfrom};
+    if (defined $sender) {
+      dbg("spf: found EnvelopeFrom '$sender' in first external Received header");	
+      $scanner->{spf_sender} = lc $sender;
+    } else {
+      dbg("spf: EnvelopeFrom not found in first external Received header");
+    }
   }
 
-  if ($sender) {
-    dbg("spf: found Envelope-From in first external Received header");
-  }
-  else {
+  if (!exists $scanner->{spf_sender}) {
     # We cannot use the env-from data, since it went through 1 or more relays 
     # since the untrusted sender and they may have rewritten it.
-    if ($scanner->{num_relays_trusted} > 0 && !$scanner->{conf}->{always_trust_envelope_sender}) {
-      dbg("spf: relayed through one or more trusted relays, cannot use header-based Envelope-From, skipping");
-      return;
-    }
-
-    # we can (apparently) use whatever the current Envelope-From was,
-    # from the Return-Path, X-Envelope-From, or whatever header.
-    # it's better to get it from Received though, as that is updated
-    # hop-by-hop.
-    $sender = $scanner->get("EnvelopeFrom:addr");
-  }
-
-  if (!$sender) {
-    dbg("spf: cannot get Envelope-From, cannot use SPF");
-    return;  # avoid setting $scanner->{sender} to undef
-  }
-
-  return $scanner->{sender} = lc $sender;
-}
-
-sub _check_spf_whitelist {
-  my ($self, $scanner) = @_;
-
-  $scanner->{spf_whitelist_from_checked} = 1;
-  $scanner->{spf_whitelist_from} = 0;
-
-  # if we've already checked for an SPF PASS and didn't get it don't waste time
-  # checking to see if the sender address is in the spf whitelist
-  if ($scanner->{spf_checked} && !$scanner->{spf_pass}) {
-    dbg("spf: whitelist_from_spf: already checked spf and didn't get pass, skipping whitelist check");
-    return;
-  }
-
-  $self->_get_sender($scanner) unless $scanner->{sender_got};
-
-  unless ($scanner->{sender}) {
-    dbg("spf: spf_whitelist_from: could not find usable envelope sender");
-    return;
-  }
-
-  $scanner->{spf_whitelist_from} = $self->_wlcheck($scanner,'whitelist_from_spf');
-  if (!$scanner->{spf_whitelist_from}) {
-    $scanner->{spf_whitelist_from} = $self->_wlcheck($scanner, 'whitelist_auth');
-  }
-
-  # if the message doesn't pass SPF validation, it can't pass an SPF whitelist
-  if ($scanner->{spf_whitelist_from}) {
-    if ($self->check_for_spf_pass($scanner)) {
-      dbg("spf: whitelist_from_spf: $scanner->{sender} is in user's WHITELIST_FROM_SPF and passed SPF check");
+    if ($scanner->{num_relays_trusted} > 0 &&
+          !$scanner->{conf}->{always_trust_envelope_sender}) {
+      dbg("spf: relayed through one or more trusted relays, ".
+            "cannot use header-based EnvelopeFrom");
     } else {
-      dbg("spf: whitelist_from_spf: $scanner->{sender} is in user's WHITELIST_FROM_SPF but failed SPF check");
-      $scanner->{spf_whitelist_from} = 0;
+      # we can (apparently) use whatever the current EnvelopeFrom was,
+      # from the Return-Path, X-Envelope-From, or whatever header.
+      # it's better to get it from Received though, as that is updated
+      # hop-by-hop.
+      my $sender = ($scanner->get("EnvelopeFrom:addr"))[0];
+      if (defined $sender) {
+        dbg("spf: found EnvelopeFrom '$sender' from header");
+        $scanner->{spf_sender} = lc $sender;
+      } else {
+        dbg("spf: EnvelopeFrom header not found");
+      }
     }
-  } else {
-    dbg("spf: whitelist_from_spf: $scanner->{sender} is not in user's WHITELIST_FROM_SPF");
+  }
+
+  if (!exists $scanner->{spf_sender}) {
+    dbg("spf: cannot get EnvelopeFrom, cannot use SPF by DNS");
   }
 }
 
-sub _check_def_spf_whitelist {
+sub _check_spf_welcomelist {
   my ($self, $scanner) = @_;
 
-  $scanner->{def_spf_whitelist_from_checked} = 1;
-  $scanner->{def_spf_whitelist_from} = 0;
+  $scanner->{spf_welcomelist_from_checked} = 1;
+  $scanner->{spf_welcomelist_from} = 0;
 
   # if we've already checked for an SPF PASS and didn't get it don't waste time
-  # checking to see if the sender address is in the spf whitelist
+  # checking to see if the sender address is in the spf welcomelist
   if ($scanner->{spf_checked} && !$scanner->{spf_pass}) {
-    dbg("spf: def_spf_whitelist_from: already checked spf and didn't get pass, skipping whitelist check");
+    dbg("spf: welcomelist_from_spf: already checked spf and didn't get pass, skipping welcomelist check");
     return;
   }
 
-  $self->_get_sender($scanner) unless $scanner->{sender_got};
-
-  unless ($scanner->{sender}) {
-    dbg("spf: def_spf_whitelist_from: could not find usable envelope sender");
+  if (!$scanner->{spf_sender}) {
+    dbg("spf: spf_welcomelist_from: no EnvelopeFrom available for welcomelist check");
     return;
   }
 
-  $scanner->{def_spf_whitelist_from} = $self->_wlcheck($scanner,'def_whitelist_from_spf');
-  if (!$scanner->{def_spf_whitelist_from}) {
-    $scanner->{def_spf_whitelist_from} = $self->_wlcheck($scanner, 'def_whitelist_auth');
-  }
+  $scanner->{spf_welcomelist_from} =
+    $self->_wlcheck($scanner, 'welcomelist_from_spf') ||
+    $self->_wlcheck($scanner, 'welcomelist_auth');
 
-  # if the message doesn't pass SPF validation, it can't pass an SPF whitelist
-  if ($scanner->{def_spf_whitelist_from}) {
+  # if the message doesn't pass SPF validation, it can't pass an SPF welcomelist
+  if ($scanner->{spf_welcomelist_from}) {
     if ($self->check_for_spf_pass($scanner)) {
-      dbg("spf: def_whitelist_from_spf: $scanner->{sender} is in DEF_WHITELIST_FROM_SPF and passed SPF check");
+      dbg("spf: welcomelist_from_spf: $scanner->{spf_sender} is in user's WELCOMELIST_FROM_SPF and passed SPF check");
     } else {
-      dbg("spf: def_whitelist_from_spf: $scanner->{sender} is in DEF_WHITELIST_FROM_SPF but failed SPF check");
-      $scanner->{def_spf_whitelist_from} = 0;
+      dbg("spf: welcomelist_from_spf: $scanner->{spf_sender} is in user's WELCOMELIST_FROM_SPF but failed SPF check");
+      $scanner->{spf_welcomelist_from} = 0;
     }
   } else {
-    dbg("spf: def_whitelist_from_spf: $scanner->{sender} is not in DEF_WHITELIST_FROM_SPF");
+    dbg("spf: welcomelist_from_spf: $scanner->{spf_sender} is not in user's WELCOMELIST_FROM_SPF");
+  }
+}
+
+sub _check_def_spf_welcomelist {
+  my ($self, $scanner) = @_;
+
+  $scanner->{def_spf_welcomelist_from_checked} = 1;
+  $scanner->{def_spf_welcomelist_from} = 0;
+
+  # if we've already checked for an SPF PASS and didn't get it don't waste time
+  # checking to see if the sender address is in the spf welcomelist
+  if ($scanner->{spf_checked} && !$scanner->{spf_pass}) {
+    dbg("spf: def_spf_welcomelist_from: already checked spf and didn't get pass, skipping welcomelist check");
+    return;
+  }
+
+  if (!$scanner->{spf_sender}) {
+    dbg("spf: def_spf_welcomelist_from: could not find usable envelope sender");
+    return;
+  }
+
+  $scanner->{def_spf_welcomelist_from} =
+    $self->_wlcheck($scanner, 'def_welcomelist_from_spf') ||
+    $self->_wlcheck($scanner, 'def_welcomelist_auth');
+
+  # if the message doesn't pass SPF validation, it can't pass an SPF welcomelist
+  if ($scanner->{def_spf_welcomelist_from}) {
+    if ($self->check_for_spf_pass($scanner)) {
+      dbg("spf: def_welcomelist_from_spf: $scanner->{spf_sender} is in DEF_WELCOMELIST_FROM_SPF and passed SPF check");
+    } else {
+      dbg("spf: def_welcomelist_from_spf: $scanner->{spf_sender} is in DEF_WELCOMELIST_FROM_SPF but failed SPF check");
+      $scanner->{def_spf_welcomelist_from} = 0;
+    }
+  } else {
+    dbg("spf: def_welcomelist_from_spf: $scanner->{spf_sender} is not in DEF_WELCOMELIST_FROM_SPF");
   }
 }
 
 sub _wlcheck {
   my ($self, $scanner, $param) = @_;
-  if (defined ($scanner->{conf}->{$param}->{$scanner->{sender}})) {
+  if (defined ($scanner->{conf}->{$param}->{$scanner->{spf_sender}})) {
     return 1;
   } else {
-    study $scanner->{sender};  # study is a no-op since perl 5.16.0
     foreach my $regexp (values %{$scanner->{conf}->{$param}}) {
-      if ($scanner->{sender} =~ qr/$regexp/i) {
+      if ($scanner->{spf_sender} =~ $regexp) {
         return 1;
       }
     }
